/*
 *      usbirboy.c  --  USB-IR-Boy board linux driver.
 *
 *	Currently only for 2.6.x kernels
 *
 *	www.sourceforge.net/projects/usbirboy/
 *
 * 	Copyright (c) 2004 Ilkka Urtamo 
 *	
 *	Used code from usb_skel and
 * 	driver skeleton from Aapo Tamminen
 *	
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/*****************************************************************************/

#include <linux/module.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <linux/delay.h>
#include <linux/usb.h>
#include <linux/poll.h>

/*
 * Version information
 */
#define DRIVER_VERSION "v0.01"
#define DRIVER_AUTHOR "Ilkka Urtamo (ilkka@urtamo.com)"
#define DRIVER_DESC "USB-IR-Boy (c)2004"
#define DRIVER_NAME "usbirboy"

#define USBIRBOY_MINOR_BASE 	240  		/* some unassigned USB minor */
#define RECEIVE_BUF_SIZE	1024		/* this must be 2^n */
#define RECEIVE_BUF_MASK	0x000003FF	/* mask for __u32 for the puffer*/
#define PULSESPACE_BIT_MASK     0x0fffffff
#define DATA_BIT_MASK           0x00ffffff

#define INC_INDEX(a) a += 1; a &= RECEIVE_BUF_MASK
#define IS_STARTBIT(a) ( a & STARTBYTE) == STARTBYTE 
#define REMOVE_STARTBIT(a) a = (a ^ STARTBYTE)
#define USBIRBOY_VENDOR_ID 	0xFFFE		// MC68HC908 , Motorola vendor ID
#define USBIRBOY_PRODUCT_ID  	0x0		// MC68HC908  USB board product ID

#define STARTBYTE		0x80		// defines if byte starts the 4 byte serie
#define __DEBUG if(debug) {
#define __END_DEBUG }



/* Strait from lirc.h adopted ioctl definitions */
/* Lets hope they dont change those*/
#define LIRC_MODE2REC(x) ((x) << 16)
#define LIRC_GET_FEATURES              _IOR('i', 0x00000000, __u32)
#define LIRC_CAN_REC_MODE2             LIRC_MODE2REC(LIRC_MODE_MODE2)
#define LIRC_GET_REC_MODE              _IOR('i', 0x00000002, __u32)
#define LIRC_MODE_MODE2                0x00000004
#define LIRC_SET_REC_MODE              _IOW('i', 0x00000012, __u32)



/* prevent races between open() and disconnect() */
static DECLARE_MUTEX (disconnect_sem);

/* define debug kernel option */
static int debug = 0;
MODULE_PARM(debug, "i");

static struct usb_driver usbirboy_driver;
static struct usb_class_driver usbirboy_class;

/* Structure to hold all of our device specific stuff */
struct usb_usbirboy {
  struct usb_device *	udev;			/* save off the usb device pointer */
  struct usb_interface *interface;		/* the interface for this device */
  unsigned char		minor;			/* the starting minor number for this device */
  unsigned char **	read_buffer;		/* the buffer to store received data packets*/
  unsigned char *	int_in_buffer;		/* the buffer to receive data from device*/
  size_t		int_in_size;		/* the size of the receive buffer */
  __u8			int_in_endpointAddr;/* the address of the interrupt in endpoint */
  __u8  		int_in_Interval;
  struct urb *		read_urb;		/* the urb used to receive data */
  unsigned char         used;                   /* true if module is open in user space */
  int			present;		/* true if the device is connected */
  struct semaphore	sem;			/* locks this structure */
  volatile int 		buff_head;
  volatile int 		buff_tail;   		/* head and tail pointers for read buffer*/
  int			packet_ptr;		/* store position in current packet */
  wait_queue_head_t	waitdata;    		/* object for userspace read to wait more data */
  wait_queue_head_t	poll_wait_packet;	/* poll/select facility for full packet ready notify */
  wait_queue_head_t 	remove_ok;		/* clear to unregister usb_device now */
};

/*internal fuction to free memory */
static inline void usbirboy_cleanup (struct usb_usbirboy *irdev)
{
  int i;
  __DEBUG
  printk("-----> %s \n", __FUNCTION__);
  __END_DEBUG

  if(irdev == NULL)
    {
      err("irdev == NULL \n");
      return;
    }
  if (irdev->int_in_buffer != NULL)
    kfree (irdev->int_in_buffer);

  if (irdev->read_urb != NULL)
    usb_free_urb (irdev->read_urb);
  if(irdev->read_buffer  != NULL)
    {
      for(i=0;i<RECEIVE_BUF_SIZE;++i)
	kfree (irdev->read_buffer[i]);
      kfree (irdev->read_buffer);
    }
  
  kfree (irdev);
  irdev = NULL;
  
  dbg("<----- %s \n", __FUNCTION__);
}
/*called from userspace, when userspace process is calling fopen for the device*/
static int usbirboy_user_open (struct inode *inode, struct file *file)
{
  /*user space has opened this device */
  /*in this version we asume we have only one device*/
  int minor = -1;
  struct usb_usbirboy *irdev;
  struct usb_interface *intf;
  int retval = -ENODEV;

  minor = iminor(inode);
  
  __DEBUG
  printk("-----> %s \n", __FUNCTION__);
  __END_DEBUG

  if(minor < 0)
    return -ENODEV;

  /*do not allow disconnecting during this function*/
  down(&disconnect_sem);

   intf = usb_find_interface(&usbirboy_driver,minor);
  if(NULL == intf)
    goto error;

  irdev = usb_get_intfdata(intf);
  if(NULL == irdev || NULL == irdev->udev || !irdev->present)
    goto error;

  __DEBUG
    printk("present = %d , used = %d\n",irdev->present,irdev->used);	  
  __END_DEBUG
  /*check if another user process is allready reading the device*/
  retval = -EBUSY;
  
  /* lock structure */
  down (&irdev->sem);
  
  if(irdev->used > 0)
    goto error_already_inuse;

  /*do actual opening*/
  irdev->used = 1;
  file->f_pos=0;
  file->private_data = irdev;

  up (&irdev->sem);
  up(&disconnect_sem);
  __DEBUG
    printk("<----- %s \n", __FUNCTION__);
  __END_DEBUG
  return 0;

 error_already_inuse:
  up (&irdev->sem);
 error:
  up(&disconnect_sem);


  return retval;
}

/*called from userspace, when userspace process is calling flose for the device*/
static int usbirboy_user_release (struct inode *inode, struct file *file)
{
  /*userspace has closed this device*/
  struct usb_usbirboy *irdev = file->private_data;
  __DEBUG
    printk("-----> %s \n", __FUNCTION__);
  __END_DEBUG

  if(irdev == NULL)
    {
      err("Error in %s : irdev is null",__FUNCTION__);
      return -ENODEV;
    } 
  __DEBUG
  printk("releasse: present = %d , used = %d\n",irdev->present,irdev->used);	  
  __END_DEBUG

  down(&irdev->sem);
  
  /*we are no longer using this device*/
  irdev->used = 0;

  if(irdev->present == 0)
    {
      
      up(&irdev->sem);
      __DEBUG
	printk("releasse: remove ok wakeup\n");	  
      __END_DEBUG
      /* device is no longer present, usb_disconnect is waiting for us */
      wake_up(&irdev->remove_ok);
      /*after this we cannot relay on irdev to be valid anymore*/
    }
  else
    up(&irdev->sem);
  __DEBUG
    printk("<----- %s \n", __FUNCTION__);
  __END_DEBUG

  return 0;

}

static int usbirboy_user_ioctl(struct inode *node,
			       struct file *filep,
			       unsigned int cmd,
			       unsigned long arg)
{
  int retval = 0;

  switch(cmd)
    {
    case LIRC_GET_FEATURES:
      retval = put_user(LIRC_CAN_REC_MODE2, (unsigned long *) arg);
      break;

    case LIRC_GET_REC_MODE:
      retval = put_user(LIRC_MODE_MODE2, (unsigned long *) arg);
      break;

    case LIRC_SET_REC_MODE:
      retval = put_user(LIRC_MODE_MODE2, (unsigned long *) arg);
      break;

    default:
      retval = -ENOIOCTLCMD;
      
    }


  return retval;

}

inline __u32 make_32b_int(unsigned char * buffer)
{
  /* We asume this buffer is 4bytes long*/

  /* each byte has 7 meaninfull bits */
  /* Combine these bits to one 32b integer*/
  __u32 data = 0;

  data |= buffer[3];
  data = data << 7;

  data |= buffer[2];
  data = data << 7;

  data |= buffer[1];
  data = data << 7;

  data |= buffer[0];

  return data;
}

static ssize_t usbirboy_user_read (struct file *file, char *buffer, size_t count, loff_t *ppos)
{
  /*user space reads the device*/
  struct usb_usbirboy *irdev = file->private_data;
  int retval = 0;
  __u32 data_int = 0;
  __DEBUG
    printk("-----> %s \n", __FUNCTION__);
  __END_DEBUG
  if (*ppos)
    return -ESPIPE;

  if(!irdev)
    {
      err("Error in %s : irdev is null",__FUNCTION__);
      return -EIO;
    }

  if (!irdev->udev)
    {
      err("Error in %s : irdev->udev is null",__FUNCTION__);
      return -EIO;
    }

  if (irdev->present == 0) 
    return -EIO;    

  /*makes no sence if userland asks less than u32 bytes*/
  if(count < sizeof(__u32))
    return 0;

  while (irdev->buff_head == irdev->buff_tail)
    {
      if (file->f_flags & O_NONBLOCK) /*userspace program do not whant us to wait for data*/
	return -EAGAIN;
      
      if (wait_event_interruptible(irdev->waitdata, (irdev->buff_head != irdev->buff_tail)))
	return -ERESTARTSYS; /* signal: tell the fs layer to handle it */

      if (irdev->present == 0) 
	return -EIO;    
    }


  down(&irdev->sem);
  
  data_int = make_32b_int(&irdev->read_buffer[irdev->buff_tail][0]);
  data_int &= PULSESPACE_BIT_MASK;

  __DEBUG
    if (data_int&0x01000000) 
      printk("pulse "); else printk("space ");
    printk("%d\n", data_int&DATA_BIT_MASK);
  __END_DEBUG

  if(copy_to_user(buffer,(char*)&data_int , sizeof(__u32)))
    retval =  -EFAULT;
  else
    retval = sizeof(__u32);
  
  INC_INDEX(irdev->buff_tail);

  up(&irdev->sem);

  return retval;

}

static unsigned int usbirboy_user_poll(struct file * file, poll_table * wait)
{
  struct usb_usbirboy *irdev = file->private_data;
  int retval=0;
  __DEBUG
    printk("%s\n", __FUNCTION__);
  __END_DEBUG

  poll_wait(file, &irdev->poll_wait_packet, wait);
  
  if(irdev->buff_head != irdev->buff_tail)
    retval =  POLLIN | POLLRDNORM;
  
 
  return retval;
}

/*This is called from interrupt handler, do the things and exit asap */
static void usbirboy_read_urb_callback (struct urb *urb, struct pt_regs *regs)
{
  struct usb_usbirboy *irdev = NULL;
  int retval=0;
  int i=0;
  int packets_ready=0;

  switch (urb->status) 
    {
    case 0:
      /* success */
      break;
    case -ECONNRESET:
      err("ECONNRESET \n");
      return;
    case -ENOENT:
      err("ENOENT  \n");
      return;
    case -ESHUTDOWN:
      err("ESHUTDOWN \n");
      return;
    default:    
      return;
    }
  irdev = urb->context;
  
  if (NULL == irdev)
    {
      err("Error in %s : irdev is null",__FUNCTION__);
      return;
    }
  if(!irdev->present)
    /* device no longer present */
    return;

  if (urb->actual_length > 0)
    {
      for(i=0;i<urb->actual_length;++i)
	{
	  if(IS_STARTBIT(irdev->int_in_buffer[i]) )
	    {
	      /*start from the beginning*/
	      if(irdev->packet_ptr != 0)
		err("new start bit before last packet was ready!\n");
	      irdev->packet_ptr = 0; 
	      irdev->read_buffer[irdev->buff_head][irdev->packet_ptr] = irdev->int_in_buffer[i];
	      REMOVE_STARTBIT(irdev->read_buffer[irdev->buff_head][irdev->packet_ptr]);	      
	    }
	  else
	    {
	      irdev->packet_ptr++;
	      if(irdev->packet_ptr<4 && irdev->packet_ptr>0)	/*just for caution*/
		irdev->read_buffer[irdev->buff_head][irdev->packet_ptr] = irdev->int_in_buffer[i];
	    }
	  
	  /*packet ready, */
	  if(irdev->packet_ptr == 3)
	    {
	      irdev->packet_ptr=0;
	      INC_INDEX(irdev->buff_head);
	      packets_ready++;
	    }
	}

      if(packets_ready>0)
	{
	  /*wake pending userland read*/
	  wake_up_interruptible(&irdev->waitdata);
	  wake_up_interruptible(&irdev->poll_wait_packet);
	  __DEBUG
	    printk(".\n");
	  __END_DEBUG
	}

    }

  // Atomic because we need to get memory emediately
  retval =  usb_submit_urb(irdev->read_urb,GFP_ATOMIC);
  
  if (retval)
    {
      err ("unable to submit urb - ret %d",retval);
      return;
    }

}

/* usb-core calles this when new usb device with 
 * id number matching one in device id table is 
 * connected
*/
static int usbirboy_usb_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
  int retval = -ENOMEM;
  size_t buffer_size;
  int i;
  struct usb_device *udev = interface_to_usbdev(interface);
  struct usb_usbirboy *irdev = NULL;
  struct usb_host_interface *iface_desc;
  struct usb_endpoint_descriptor *endpoint;
  struct usb_device_descriptor *irdescriptor  = NULL;
  __DEBUG
    printk("-----> %s \n", __FUNCTION__);
    printk( "dev Product: 0x%x\n",udev->descriptor.idProduct);
    printk( "dev venrod: 0x%x\n",udev->descriptor.idVendor);
  __END_DEBUG


  /* See if the device offered us matches what we can accept */
  if ((udev->descriptor.idVendor != USBIRBOY_VENDOR_ID) ||
      (udev->descriptor.idProduct != USBIRBOY_PRODUCT_ID)) 
    {
      return -ENODEV;
    }
  
  irdev = kmalloc (sizeof(*irdev), GFP_KERNEL);
  if (irdev == NULL) 
    {
      err ("Out of memory");
      goto error;
    }
  memset (irdev, 0x00, sizeof (*irdev));
  
  
  irdev->read_buffer =(unsigned char **) kmalloc (
						  RECEIVE_BUF_SIZE*sizeof(unsigned char**),
						  GFP_KERNEL);
  for(i=0;i<RECEIVE_BUF_SIZE;++i)
    {
      irdev->read_buffer[i] =(unsigned char*) kmalloc (4*sizeof(unsigned char*), GFP_KERNEL);
      memset (irdev->read_buffer[i] , 0x00, 4*sizeof(unsigned char*));
    }
  __DEBUG
    printk("init_MUTEX dev->sem \n");	
  __END_DEBUG

  init_MUTEX (&irdev->sem);
  
  irdev->udev = udev;
  irdev->interface = interface;
  irdescriptor  = &udev->descriptor;
  /* seach for correct endpoints and initialize them*/
  iface_desc = &interface->altsetting[0];
  __DEBUG
    printk( "bNumEndpoints: %d\n", iface_desc->desc.bNumEndpoints);
  __END_DEBUG
  
  /*Just pick up fist 2 endpoints*/
  for(i = 0;i<iface_desc->desc.bNumEndpoints; ++i )
    {
      int maxp =0 ;
      unsigned int pipe;
      endpoint = &iface_desc->endpoint[i].desc;
      
      if (!irdev->int_in_endpointAddr &&
	  (endpoint->bEndpointAddress & USB_DIR_IN) && // with mask == 1
	  ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
	   == USB_ENDPOINT_XFER_INT))
	{
	  __DEBUG
	    printk("Using endpoint %d for reading, addr=%x attr=%x\n",
	      i,
	      endpoint->bEndpointAddress,
	      endpoint->bmAttributes);
	  __END_DEBUG
	  buffer_size = endpoint->wMaxPacketSize;
	  irdev->int_in_size = buffer_size;
	  irdev->int_in_Interval = endpoint->bInterval;
	  irdev->int_in_endpointAddr = endpoint->bEndpointAddress;
	  irdev->int_in_buffer = kmalloc (buffer_size, GFP_KERNEL);
	  
	  if (!irdev->int_in_buffer) 
	    {
	      err("Couldn't allocate int_in_buffer");
	      goto error;
	    }
	  memset (irdev->int_in_buffer, 0x00, sizeof (*irdev->int_in_buffer));
	  
	  irdev->read_urb = usb_alloc_urb(0, GFP_KERNEL);	
	  if (!irdev->read_urb)
	    {
	      err("No free urbs available for read_urb");
	      goto error;
	    }
	  pipe = usb_rcvintpipe(udev, endpoint->bEndpointAddress);
	  maxp = usb_maxpacket(udev, pipe, usb_pipeout(pipe));
	  __DEBUG
	  printk("maxpr=%d\n",buffer_size);
	  __END_DEBUG

	  usb_fill_int_urb(irdev->read_urb, 
			   udev,
			   pipe,		/*receiving pipe*/
			   irdev->int_in_buffer, 
			   maxp,
			   usbirboy_read_urb_callback,
			   irdev,
			   irdev->int_in_Interval);	
	  
	}
      
    }


  init_waitqueue_head (&irdev->waitdata);
  init_waitqueue_head (&irdev->remove_ok);
  init_waitqueue_head (&irdev->poll_wait_packet);

  /* allow device read */
  irdev->present = 1;
  
  /*reset read buffer*/
  irdev->buff_head = 0;
  irdev->buff_tail = 0;
  /*device is not used in userspace at this point*/
  irdev->used = 0;
  irdev->packet_ptr=0;

  /* we can register the device now, as it is ready */
  usb_set_intfdata (interface, irdev);						
  
  retval = usb_register_dev (interface, &usbirboy_class);		
  if (retval)
    {
      /* something prevented us from registering this driver */
      err ("Not able to get a minor for this device.");
      usb_set_intfdata (interface, NULL);
      goto error;
    }

  irdev->minor = interface->minor;  

  retval =  usb_submit_urb(irdev->read_urb,GFP_KERNEL);  		
  if (retval)
    {
      err ("unable to submit urb ");
      goto error;
    }
  

  
  /* let the user know what node this device is now attached to */
  info("USB-IR-Boy device attached to usbirboy with minor %d\n", irdev->minor);
  __DEBUG
    printk("<----- %s \n", __FUNCTION__);
  __END_DEBUG
  return retval;
  
 error:
  usbirboy_cleanup (irdev);
  
  dbg("error: <----- %s \n", __FUNCTION__);
  return retval;
}

/* This is called by usb-core whenever device is disconnected */
static void usbirboy_usb_disconnect(struct usb_interface *interface)
{
	struct usb_usbirboy *usbirdev = NULL;
	int minor = -1;
	__DEBUG
	  printk("-----> %s \n", __FUNCTION__);
	__END_DEBUG

	/* prevent open to get thrue because we are disconnecting */
	down (&disconnect_sem);
	
	usbirdev = usb_get_intfdata (interface);
	usb_set_intfdata (interface, NULL);


	/*lock dev structure*/
	down (&usbirdev->sem);
	minor = usbirdev->minor;

	/* terminate an ongoing read */
	__DEBUG
	  printk("disconnect: unlinking urb...\n");
	__END_DEBUG
	/*this will end all further interrupts from device*/
	usb_unlink_urb (usbirdev->read_urb);
	
	/* prevent device read  */
	usbirdev->present = 0;

	if(usbirdev->used == 1)
	  {
	    /*driver is open to user space */
	    __DEBUG
	      printk("disconnect: used == 1\n");	  
	    __END_DEBUG
	    /*unlock structure to let read finish and release */
	    up (&usbirdev->sem); 

	    /* wake up pending read */
	    __DEBUG
	    printk("disconnect: wake up wait data\n");	  
	    __END_DEBUG
	    wake_up (&usbirdev->waitdata);
	    
	    /*wait untill userspace releases the driver*/
	    __DEBUG
	      printk("disconnect: waiting remove_ok\n");	  
	    __END_DEBUG

	    sleep_on(&usbirdev->remove_ok);

	    __DEBUG
	      printk("disconnect: got remove_ok\n");	  
	    __END_DEBUG
	    /*continue with locked structure*/
	    down (&usbirdev->sem);
	  }
	
	/**** clean up ****/
	
	/* give back our minor */
	usb_deregister_dev (interface, &usbirboy_class);
	
	/* if device is open to userspace, release will eventually clean this up*/
	__DEBUG
	  printk("present = %d , used = %d\n",usbirdev->present,usbirdev->used);
	__END_DEBUG
	
	if (usbirdev->used == 1)
	  err("Error: driver still used by userspace.. cleaning anyway");

	up (&usbirdev->sem);
	usbirboy_cleanup (usbirdev);
	up (&disconnect_sem);

	info( "USB-IR-Boy device disconnected. Minor: %d\n",minor);
	__DEBUG
	printk("<----- %s \n", __FUNCTION__);
	__END_DEBUG
}

/*called by usb-core during insmod or modprobe */
static int __init usbirboy_module_init (void)
{
  int result;
  __DEBUG
    printk("Loaded with debug option, flooding debug...\n");
    printk("-----> %s \n", __FUNCTION__);
  __END_DEBUG
  if(RECEIVE_BUF_SIZE % 4 != 0)
    {
      err("usbirboy has been compiled with RECEIVE_BUF_SIZE not counts of 4bytes!");
      return -1;
    }    
  
  /* register this driver with the USB subsystem */
  result = usb_register(&usbirboy_driver);			
  if (result < 0)
    {
      err("usb_register failed for the " __FILE__
	  " driver. Error number %d", result);
      return -1;
    }
   
  info( DRIVER_DESC ":" DRIVER_VERSION);

  __DEBUG
  printk("<----- %s \n", __FUNCTION__);
  __END_DEBUG

  return 0;

}

/*called by usb-core during rmmod*/
static void __exit usbirboy_module_exit (void)
{
  __DEBUG
    printk("----->  %s \n", __FUNCTION__);
  __END_DEBUG
	usb_deregister (&usbirboy_driver);
	
  info( "usbirboy_driver deregistered...\n");
  __DEBUG
    printk("<----- %s \n", __FUNCTION__);
  __END_DEBUG
}

/*talbe describing product and vendor id's we use this driver for */
static struct usb_device_id usbirboy_device_table [] = {
    { USB_DEVICE(USBIRBOY_VENDOR_ID,
      USBIRBOY_PRODUCT_ID) },
    { }                      /* Terminating entry */
};

/*defines user space function bodies*/
static struct file_operations usbirboy_fops = {
	.owner =	THIS_MODULE,
	.read =		usbirboy_user_read,
	.ioctl = 	usbirboy_user_ioctl,
	.open =		usbirboy_user_open,
	.release =	usbirboy_user_release,
	.poll = 	usbirboy_user_poll,
};

/*descriptor for devfs */
static struct usb_class_driver usbirboy_class = {
	.name =		DRIVER_NAME,
	.fops =		&usbirboy_fops,
	.mode =		S_IFCHR | S_IRUSR | S_IRGRP | S_IROTH,
	.minor_base =	USBIRBOY_MINOR_BASE,
};

/*driver info for kernel*/
static struct usb_driver usbirboy_driver = {
        .owner =        THIS_MODULE,
        .name =         DRIVER_NAME,
        .id_table =     usbirboy_device_table,
        .probe =        usbirboy_usb_probe,
        .disconnect =   usbirboy_usb_disconnect,

};
/* -------------------------- Ending ------------------------------------ */

MODULE_DEVICE_TABLE (usb, usbirboy_device_table);

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

module_init (usbirboy_module_init);
module_exit (usbirboy_module_exit);


/* --------------------------------------------------------------------- */


